"""The plotting sub-package contains plot methods and tools for Pastas.

.. currentmodule:: pastas.plotting

.. autosummary::
    :toctree: ./generated
    :nosignatures:

    plots
    modelplots.Plotting
    plotly.Plotly
    bokeh.Bokeh
    modelcompare
    plotutil

"""
